import cca_zoo.data
import cca_zoo.models
import cca_zoo.model_selection

try:
    import cca_zoo.deepmodels
except ModuleNotFoundError:
    pass
try:
    import cca_zoo.probabilisticmodels
except ModuleNotFoundError:
    pass
