from .rcca import rCCA, CCA, PLS
from .mcca import MCCA, KCCA
from .gcca import GCCA, KGCCA
from .iterative import (
    _Iterative,
    PLS_ALS,
    PMD,
    CCA_ALS,
    ElasticCCA,
    ParkhomenkoCCA,
    SCCA,
    SCCA_ADMM,
    SpanCCA,
    SWCCA,
)
from .ncca import NCCA
from .partialcca import PartialCCA
from .tcca import TCCA, KTCCA
